# Neo Luminosity & Lightness Mask  Maker 8 & 16bit Version
# Ver. 0.02 reactive version
# add tranceparency factor
# add luminosity mask
# add Nega mask (2022.5)
# (c)2021-2 Ohnishi, Yasuo

from ij import IJ, ImagePlus
from ij import WindowManager as WM
from ij.gui import GenericDialog, ImagePanel
from ij.io import DirectoryChooser, OpenDialog
# from ij.gui import GenericDialog
from ij import plugin
from ij.plugin import ChannelSplitter, Duplicator
from ij.plugin import RGBStackMerge, RGBStackConverter
from ij.plugin.filter import ThresholdToSelection
from ij.process import ImageConverter, ImageProcessor
from ij.process import ByteProcessor
from loci.plugins import BF
from java.awt.event import AdjustmentListener, ItemListener
import os
from org.python.core import codecs
codecs.setDefaultEncoding('utf-8')

#---Convert Value in 8, 16, 32bit mode
def convValue(Value, Mode):
#     Converting Luminance Value each in 8, 16, 32bit mode
        if Mode == "16-bit":
                Value = Value * 256
        elif Mode == "32-bit":
                Value = Value * 256 * 256
        else:
                Value = Value
        return Value

#--------------------------------------------
def luminosity_mask_maker(imp1, LTh, Mode, NegaInvert, Show, TrFac, chName):
# imp1: Grayscale image
# LTh: Zone lower Threshold
# NegaInvert: Nega invert True or False
# TrFac: Transparency Factor
# Close Windows
#    IJ.run(imp1, "Conversions...", "weighted")
    if Show == "Yes":
        imp1.show()
# Make Grayscale Mask Image Files 
    V1 = str(LTh) # for threshold expression in file name
# Target (Yellow) Mask Threshold
    V2 = str(convValue(abs(LTh), Mode)) # actual Yellow mask Threshold
    IJ.run(imp1, "Subtract...","value=" + V2)
    IJ.run(imp1, "Multiply...", "value=" + str(4 * TrFac)) # making mask more transparent (x 4)
    if NegaInvert == True: # for making inverted Negative mask
        """
        IJ.run(imp1, "Multiply...", "value=" + str(4 / TrFac)) # making mask more transparent (x 4 for Nega)
        imp1.getProcessor().invert()
        """
        IJ.run(imp1, "Invert","")
        V1 = V1 + "_Nega"
        V = "_Nega_x "+ str(TrFac)
    else: # for making positive mask
        V1 = V1 + "_Posi"
        V = "_Posi_x" + str(TrFac)
    impGraymask = imp1
    if Show == "Yes":
        impGraymask.show()  # Show the subtracted image in window
    IJ.saveAsTiff(impGraymask, folder + filename + "_" + chName + "_Mask_Thr_".decode('utf-8') + V1.decode('utf-8') + "_x4x" + str(TrFac) + ". tif".decode('utf-8')) # Save the Subtracted image x 4 x transparency factor
#   save grayscalee mask
    return impGraymask, V

#-------------------------------------------
def make_binary_mask(impZone, zoneLTh, zoneHTh, Mode, impGraymask, Show, chName):
# Attention!! impZone is absolutery 8bit!
    zoneLTh = 0 # for Luminosity mask zone low threshold must be always 0!
    ip = impZone.getProcessor()
    mask = ByteProcessor(ip.width, ip.height)
    mask.setValue(255)
    if Show == "Yes":
        impZone.show()
# make Zone mask
    impZone.getProcessor().setThreshold(zoneLTh, zoneHTh, ImageProcessor.NO_LUT_UPDATE)
    print "Zone L & H", zoneLTh, zoneHTh
    roi = ThresholdToSelection.run(impZone)
    mask.setRoi(roi)
    mask.fill(roi)
    try:
        maskimp = ImagePlus("Mask", mask) 
#        only 8bit Zone mask made from ROI: Zone mask: maskimp
    except: # exception process for ROI missing
        IJ.error("Selected ROI (Region of Interest) is missing.")
        if Show == "Yes":
            impGraymask.hide()
            impZone.hide()
        impGraymask.close()
        impZone.close()
        return
    if Show == "Yes":
        maskimp.show() # show Zone mask
#    IJ.saveAsTiff(maskimp, folder + filename +"_tmp_" + chName + "_Zone" + str(int(zoneLTh)) + "-" + str(int(zoneHTh)) + "_Mask.tif")

# deepen Zone Mask bit depth  (8bit to 16 or 32 bit)      
    if Mode == "16-bit":
        ImageConverter(maskimp).convertToGray16()
    elif Mode == "32-bit":
        ImageConverter(maskimp).convertToGray32()
    IJ.run(maskimp, "Multiply...", "value=" + str(convValue(1, Mode)))
#   after deepen bit depth, Save Zone Mask
    IJ.saveAsTiff(maskimp, folder + filename +"_" + chName + "_Zone" + str(int(zoneLTh)) + "-" + str(int(zoneHTh)) + "_Mask.tif")
    return maskimp

#----------------------------------------------
### RGB Manipulation 
def RGBmanipulation(imp0, folder,filename, baseCh, parm, colorCh, stackSize):
#  input Parameters
    if parm is not None: # if parm is not null, then get paraneters...
        Mode, YTh, blInvert, zoneLTh, zoneHTh, TrFac, Show = parm 
#     Mode: Image Format 16-bit or 8-bit
#     YTh: yellow(or blue) value threshold of mask
#     zoneLTh: lower threshold of mask
#     zoneHTh: upper threshold of mask
#     blinvert: make blue mask or not
#     Show: show working windows or not
# Windows Showing Mode
        imp0.show()
    impsCh = ChannelSplitter.split(imp0) #Channel Splitting of imp0
    for i in impsCh: # it needs to convert Grayscale image
        if Mode == "8-bit":
            ImageConverter(i).convertToGray8()
        elif Mode == "16-bit":
            ImageConverter(i).convertToGray16()
        else:
            ImageConverter(i).convertToGray32()
    if stackSize !=1: # if imp0 is not monochrome
        """ R ch.""" 
        impR =  impsCh[0]
        """ G ch.""" 
        impG =  impsCh[1]
        """ B ch. & B x 0.3""" 
        impB =  impsCh[2]
# Show Windows
#    imp0.close()
    if Show == "Yes":
        if stackSize !=1: # if imp0 is not monochrome
            impR.show()
            impG.show()
            impB.show()
        imp0.hide()
    if colorCh == "Blue":
        imp1 = Duplicator().run(impB)
        chName0 = "B"
    elif colorCh == "Red":
        imp1 = Duplicator().run(impR)
        chName0 = "R"
    elif colorCh == "Green":
        imp1 = Duplicator().run(impG)
        chName0 = "G"
    else: #Lightness or Luminosity
        imp1 = Duplicator().run(imp0)
        chName0 = "L"
        if colorCh == "Luminosity":
            chName0 = "Lm"
            IJ.run(imp1, "Conversions...", "weighted")
        if Mode == "16-bit":
            ImageConverter(imp1).convertToGray16()
        elif Mode == "8-bit":
            ImageConverter(imp1).convertToGray8()
        imp1.getProcessor().setLut(None)

#-----------------------------------
# Make Luminosity Mask
    impYmask, V = luminosity_mask_maker(imp1, zoneLTh, Mode, blInvert, Show, TrFac, chName0)


#############################################################
# select base channel for deciding mask upper and lower threshold
    if baseCh == "Luminosity":
        impZone = Duplicator().run(imp0)
        IJ.run(impZone, "Conversions...", "weighted")
        chName1 ="Lm"
        ImageConverter(impZone).convertToRGB()
    elif baseCh == "Lightness":
        impZone = Duplicator().run(imp0)
        chName1 ="L"
        ImageConverter(impZone).convertToRGB()
    elif baseCh == "Blue":
        impZone = Duplicator().run(impB)
        chName1 ="B"
    elif baseCh == "Red":
        impZone = Duplicator().run(impR)
        chName1 ="R"
    elif baseCh == "Green":
        impZone = Duplicator().run(impG)
        chName1 ="G"
# if imp0 is not monochrome
    if stackSize != 1:
        impR.close()
        impG.close()
        impB.close()
    imp0.close()
#-----------------------------------------------
# impZone for Grayscale must be absolutely 8 bit!!
    ImageConverter(impZone).setDoScaling(True)
#           it is needs for 16bit image to down convert to 8bit
    ImageConverter(impZone).convertToGray8()
# call make_binary_mask
    maskimp = make_binary_mask(impZone, zoneLTh, zoneHTh, Mode, impYmask, Show, chName1)
# maskimp may be 8bit or 16bit depending on original image bit
#--------------------------------------------------
# bind Zone mask and Yellow / blue mask
    ic = plugin.ImageCalculator()
    maskBindimp = ic.run("Min created", impYmask, maskimp) # bind masks
    if Show == "Yes":
        maskBindimp.show()
    IJ.saveAsTiff(maskBindimp, folder + filename +"_" + chName0 + "_Zone" + str(int(zoneLTh)) + "-" + str(int(zoneHTh)) + "_" + chName1 + V + "_Mask_Binded.tif")
#     save binded mask
    if Show == "Yes":
        impYmask.hide()
        impZone.hide()
        maskimp.hide()
        maskBindimp.hide()
    impYmask.close()
    maskimp.close()
    maskBindimp.close()
    impZone.close()
#   End of def RGBmanipulation 
#----------------------------------------------

class ThresholdPreviewer(AdjustmentListener, ItemListener): # observe changes of dialog items
    def __init__(self, imp, slider_upper, slider_lower, checkbox_prv):
        """
             imp: an ImagePlus
             slider: a java.awt.Scrollbar UI element
             preview_checkbox: a java.awt.Checkbox controlling whether to
                                                 dynamically update the ImagePlus as the
                                                 scrollbar is updated, or not.
        """
        self.imp = imp
        self.original_ip = imp.getProcessor().duplicate() # store a copy
        self.imp_org = Duplicator().run(imp0)
        self.slider_upper = slider_upper
        self.slider_lower = slider_lower
        self.checkbox_prv = checkbox_prv

    def adjustmentValueChanged(self, event):
        """ event: an AdjustmentEvent with data on the state of the scroll bar. """
        preview = self.checkbox_prv.getState()
        if preview:
            if event.getValueIsAdjusting():
                return # there are more scrollbar adjustment events queued already
            self.threshold()

    def itemStateChanged(self, event):
        """ event: an ItemEvent with data on what happened to the checkbox. """
        if event.getStateChange() == event.SELECTED:
            self.threshold()
    
    def reset(self):
        """ Restore the original ImageProcessor """
        self.imp.setProcessor(self.original_ip)
    
    def threshold(self):
        """ Execute the in-place making binary image of the ImagePlus. """
        ip = self.imp.getProcessor()
        mask = ip.createProcessor(ip.width, ip.height)
        mask.BLACK_AND_WHITE_LUT
        mask.setValue(convValue(255, Mode))
        threshold_upper = self.slider_upper.getValue()
        threshold_lower = self.slider_lower.getValue()
        if threshold_upper < threshold_lower:
            threshold_upper = threshold_lower
        self.original_ip.setThreshold(threshold_lower, threshold_upper, ImageProcessor.OVER_UNDER_LUT)
        roi = ThresholdToSelection().convert(self.original_ip) # make ROI from threshold
        mask.setRoi(roi)
        mask.fill(roi)
        self.imp.setProcessor(mask)
        self.imp.getProcessor().setLut (None)

def thresholdImageUI(imp, Mode, C0, C1):
    ImageConverter(imp).convertToGray8()
#    imp = WM.getCurrentImage()
#    imp.plotHistogram()
    WM.getActiveWindow().toFront()
#    imppanel = ImagePanel(imp)
    gd = GenericDialog(C0 +" (" + C1 + ") Mask Maker / input threshold parameters")
#    if C0 != "Luminosity Positive":
#        gd.addNumericField(C0 +" (" + C1 + ") Mask Threshold \n (Signed relative value)", 10, 0)
#    else:
#        gd.addNumericField("This field is not available for Luminosity Mask", 0, 0)
    gd.addCheckbox("Making " + C1 + " Nega Mask", True)
    gd.addSlider("Upper Threshold of Mask", 0, 255, 128)
    gd.addSlider("Lower Threshold of Mask", 0, 255, 128)
    gd.addCheckbox("Preview", True)
    gd.addNumericField("Mask Transparency Factor", 1.0, 1)
    showWin = ["Yes","No"]
    gd.addChoice("Show working windows?", showWin, showWin[0])

    # The UI elements for the above two inputs
    slider_upper = gd.getSliders().get(0) # the 1st
    slider_lower = gd.getSliders().get(1) # the 2nd
    checkbox_Bmask = gd.getCheckboxes().get(0) #  chkbox for blue mask
    checkbox_prv = gd.getCheckboxes().get(1) # chkbox for preview
#    nfield_maskThr = gd.getNumericFields().get(0) # 1st start
    nfield_TrFac = gd.getNumericFields().get(0) # Transparency Factor
    choice_showWin = gd.getChoices().get(0)


    if not imp:
        print "Open an image first!"
        return

    previewer = ThresholdPreviewer(imp, slider_upper, slider_lower, checkbox_prv)
    slider_upper.addAdjustmentListener(previewer) # set listener for upper slider
    slider_lower.addAdjustmentListener(previewer) # set listener for lower slider
    checkbox_prv.addItemListener(previewer)  # set listener for preview checkbox
    
    gd.showDialog()

    if gd.wasCanceled():
        previewer.reset()
        Mode = "Stop"
        print "User canceled dialog!"
#    else:
#        previewer.threshold()
#    parm_Mode   = "Auto"
    YTh =  0
    blInvert = gd.getNextBoolean()
    zoneHTh =  gd.getNextNumber()
    zoneLTh =  gd.getNextNumber()
    TrFac =  gd.getNextNumber()
    Show =  gd.getNextChoice()
    imp.hide()
    imp.close()
    return Mode, YTh, blInvert, zoneLTh, zoneHTh, TrFac, str(Show)


def selectChannel(): # dialog of selecting target channel for mask upper & lower threshold
    gd = GenericDialog("Lightness Mask Maker / Target Channel Selection")
    scolorCh = ["Lightness", "Luminosity", "Red","Green","Blue"]
    gd.addChoice("Select Target Channel for Mask", scolorCh, scolorCh[0])
    sbaseCh = ["Lightness", "Luminosity","Red","Green","Blue"]
    gd.addChoice("Select Target Channel for deciding upper & lower threshold", sbaseCh, sbaseCh[0])
    gd.showDialog()

    colorCh =  gd.getNextChoice()
    baseCh =  gd.getNextChoice()
    if gd.wasCanceled():
        colorCh =  "None"
        baseCh =  "None"

    return str(baseCh),  str(colorCh)

#----------main----------
window = WM.getActiveWindow()
Mode = "Auto"
if not window:
    od = OpenDialog("Choose a file")
    folder = od.getDirectory()
    filename = od.getFileName()
    if filename:
        path = folder + filename
        IJ.log(path)
        IJ.log(folder)
        IJ.log(filename)
# Always Open file with Bio-format Plugin
        imps = BF.openImagePlus(folder + filename) # Open file with Bio-format Plugin
        imp0 = imps[0] #imp0: Original Image 
        imp0.show()
    else:
        Mode = "Stop"
else:
    imp0 = IJ.getImage()
    filename = imp0.getTitle()
    folder = IJ.getDirectory("image") # get image's original directory
    folder = IJ.getDirectory("Output_directory")
if Mode == "Auto":
    bitdepth = imp0.getBitDepth()
    if bitdepth == 8:
            Mode = "8-bit"
    elif bitdepth ==16:
            Mode = "16-bit"
    elif bitdepth ==32:
            Mode = "32-bit"
    else:
            Mode = "16-bit"
stackSize = imp0.getStackSize() # imp0 is color or monochrome?
if Mode <> "Stop": # if user did not cancel execute process
    impprv = Duplicator().run(imp0) # impprv: preview imp
    scale = 1200.0 / impprv.getWidth() # preview image scaling value (width = 1200px fixed)
    new_height = impprv.getHeight() * scale # preview image height
    new_height = int(new_height)
    tmpimpprv = impprv.resize(1200, new_height, "none")
    ImageConverter(tmpimpprv).convertToRGB() # convert to 8bit RGB
    impprvsCh = ChannelSplitter.split(tmpimpprv)
    
    if stackSize == 1: #imp0 is monochrome
        baseCh = "Luminosity"
        colorCh = "Luminosity"
    else:  #imp0 is not monochrome
        baseCh, colorCh = selectChannel()
        """ select target channel for deciding upper & lower threshold """
    if colorCh == "Blue":
        C0 = "Blue"
        C1 = "Blue inverted"
    elif colorCh == "Red":
        C0 = "Red"
        C1 = "Red inverted"
    elif colorCh == "Green":
        C0 = "Green"
        C1 = "Green inverted"
    elif colorCh == "Luminosity":
        C0 = "Luminosity Positive"
        C1 = "Luminosity Negative"
    elif colorCh == "Lightness":
        C0 = "Lightness Positive"
        C1 = "Lightness Negative"

    if baseCh != "None": # if process is not canceled
        imp0.hide()
        if baseCh == "Blue":
            impprvsCh[2].show()
            parm = thresholdImageUI(impprvsCh[2], Mode, C0, C1)
        elif baseCh == "Red":
            impprvsCh[0].show()
            parm = thresholdImageUI(impprvsCh[0], Mode, C0, C1)
        elif baseCh == "Green":
            impprvsCh[1].show()
            parm = thresholdImageUI(impprvsCh[1], Mode, C0, C1)
        else:
    #        ImageConverter(impprv).convertToRGB()
            ImageConverter(tmpimpprv).convertToGray8()
            tmpimpprv.show()
            parm = thresholdImageUI(tmpimpprv, Mode, C0, C1)
    else: # if process is  canceled
        parm = []
        parm.append("Stop")
        imp0.hide()
        imp0.close()
    tmpimpprv.hide()
    impprvsCh[0].hide()
    impprvsCh[1].hide()
    impprvsCh[2].hide()
    tmpimpprv.close()
    impprvsCh[0].close()
    impprvsCh[1].close()
    impprvsCh[2].close()
    
    if parm[0] <> "Stop":
        RGBmanipulation(imp0, folder, filename, baseCh, parm, colorCh, stackSize)
else:
    IJ.log("User canceled")
IJ.log("Finish")

